import numpy as np

from stimupy.components.edges import cornsweet as cornsweet_edge

__all__ = [
    "cornsweet",
]


def cornsweet(
    visual_size=None,
    ppd=None,
    shape=None,
    ramp_width=None,
    rotation=0.0,
    intensity_edges=(0.0, 1.0),
    intensity_plateau=0.5,
    exponent=2.75,
):
    """Draw rectangular Cornsweet edge stimulus.

    The 2D luminance profile of the stimulus is defined as
    Left side:
    v = vtarget + (1 - X / w) ** a * (intensity_max-vtarget) for the ramp and v = vtarget beyond.
    Right side:
    v = vtarget - (1 - X / w) ** a * (intensity_min-vtarget) for the ramp and v = vtarget beyond.
    X is the distance to the edge, w is the width of the ramp, a is a variable
    determining the steepness of the ramp, vtarget is the luminance of the targets and
    intensity_max/intensity_min are the max/min luminances.

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    ramp_width : float
        width of luminance ramp in degrees of visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_edges : (float, float)
        intensity of edges
    intensity_plateau : float
        intensity value of plateau
    exponent : float
        determines steepness of ramp (default is 2.75. 1 would be linear)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Boyaci, H., Fang, F., Murray, S.O., Kersten, D. (2007).
        Responses to lightness variations in early human visual cortex.
        Current Biology 17, 989-993.
        https://doi.org/10.1016/j.cub.2007.05.005
    Cornsweet, T. (1970).
        Visual perception. Academic press.
        https://doi.org/10.1016/B978-0-12-189750-5.X5001-5
    """

    stim = cornsweet_edge(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        ramp_width=ramp_width,
        rotation=rotation,
        intensity_edges=intensity_edges,
        intensity_plateau=intensity_plateau,
        exponent=exponent,
    )

    # Generate the target mask
    mask = np.where(stim["d1"] == 1, stim["edge_mask"], 0) + np.where(
        stim["d2"] == 1, stim["edge_mask"], 0
    )
    stim["target_mask"] = mask.astype(int)
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {"visual_size": 15, "ppd": 30}
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "cornsweet": cornsweet(**default_params, ramp_width=3)
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import copy

import numpy as np

import stimupy

np.random.seed(1)

__all__ = ["logo"]


def logo(
    ppd=128,
    intensity_background=0.5,
):
    """Generate stimupy logo

    Parameters
    ----------
    ppd : Number (default: 128)
        pixels per degree along the axis of grating
    intensity_background : Number
        intensity value of background (default: 0.5)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        and additional keys containing stimulus parameters
    """

    # Parameters
    res = {"visual_size": 11, "ppd": ppd}
    radius = 5

    segments = stimupy.waves.square_angular(**res, n_segments=8, rotation=-(360 / 16))
    circle = stimupy.components.shapes.circle(**res, radius=radius, intensity_background=0)
    pie = copy.deepcopy(segments)

    pie.update(
        img=np.where(circle["circle_mask"], segments["img"], intensity_background),
        mask=np.where(circle["circle_mask"], segments["grating_mask"], 0),
        radius=circle["radius"],
    )

    # Create stimuli for pies
    grating = stimupy.stimuli.waves.sine_linear(**res, frequency=1, rotation=(360 / 8))
    dotted = stimupy.stimuli.sbcs.dotted(
        **res, n_dots=10, dot_radius=0.25, distance=0.1, target_shape=2
    )
    dotted = stimupy.utils.roll_dict(dotted, shift=-dotted["shape"][0] / 4, axes=0)
    grid = stimupy.stimuli.hermanns.grid(**res, element_size=(0.5, 0.5, 0.07))
    whites = stimupy.stimuli.whites.radial(**res, target_indices=(7, 14), frequency=2)
    plaid = stimupy.stimuli.plaids.sine_waves(
        grating_parameters1={**res, "frequency": 2},
        grating_parameters2={**res, "frequency": 1, "rotation": 45},
    )
    mondrian = stimupy.mondrians.corrugated_mondrian(
        **res, intensities=np.random.rand(10, 10), depths=(0, 0, 0, 0, 0, 0, 0, 1, -1, 0)
    )
    noise = stimupy.noises.naturals.pink(**res)

    # Fill segments
    pie["img"] = np.where(pie["mask"] == 1, intensity_background, pie["img"])
    pie["img"] = np.where(pie["mask"] == 2, grating["img"], pie["img"])
    pie["img"] = np.where(pie["mask"] == 3, dotted["img"], pie["img"])
    pie["img"] = np.where(pie["mask"] == 4, grid["img"], pie["img"])
    pie["img"] = np.where(pie["mask"] == 5, whites["img"], pie["img"])
    pie["img"] = np.where(pie["mask"] == 6, plaid["img"], pie["img"])
    pie["img"] = np.where(pie["mask"] == 7, mondrian["img"], pie["img"])
    pie["img"] = np.where(pie["mask"] == 8, noise["img"], pie["img"])

    # Create eye components
    lens = stimupy.components.lines.ellipse(**res, radius=radius, line_width=0.02)
    pupil = stimupy.components.shapes.ellipse(**res, radius=(radius - 1.8, radius - 0.6))
    pupilmask = stimupy.components.shapes.rectangle(
        **res, rectangle_size=(11, 3), rectangle_position=(0, 0)
    )
    pupil["img"] = pupil["img"] * pupilmask["rectangle_mask"]
    pupil["ellipse_mask"] = (pupil["ellipse_mask"] * pupilmask["rectangle_mask"]).astype(int)
    pupil = stimupy.utils.roll_dict(pupil, shift=-pupil["shape"][1] / 5, axes=1)
    pupil["ellipse_mask"] = np.where(
        (pupil["ellipse_mask"] * segments["grating_mask"] * circle["circle_mask"]) == 1,
        pupil["ellipse_mask"],
        0,
    )

    pie["img"] = np.where((segments["grating_mask"] * lens["line_mask"]) == 1, 0, pie["img"])
    pie["img"] = np.where(pupil["ellipse_mask"] == 1, 0, pie["img"])

    return pie


def qrcode(ppd=128):
    import qrcode
    from PIL import Image, ImageDraw
    from qrcode.image.styledpil import StyledPilImage

    arr = np.array(logo(ppd, intensity_background=1)["img"])
    arr *= 255

    imsize = arr.shape
    im_logo = Image.new("L", (imsize[1], imsize[0]))
    im_logo.putdata(arr.flatten())

    qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_H)
    qr.add_data("https://stimupy.readthedocs.io/")
    img = qr.make_image(
        image_factory=StyledPilImage,
        module_drawer=qrcode.image.styles.moduledrawers.CircleModuleDrawer(),
        # eye_drawer=qrcode.image.styles.moduledrawers.CircleModuleDrawer(),
        embeded_image=im_logo,
    )

    # Make "eyes" circular
    eye_size = 70  # default
    quiet_zone = 40  # default

    eyes_mask = Image.new("L", img.size, 0)

    eye_border = quiet_zone + eye_size

    draw = ImageDraw.Draw(eyes_mask)
    draw.rectangle((quiet_zone, quiet_zone, eye_border, eye_border), fill=255)
    draw.rectangle(
        (img.size[0] - eye_border, quiet_zone, img.size[0] - quiet_zone, eye_border), fill=255
    )
    draw.rectangle(
        (quiet_zone, img.size[0] - eye_border, eye_border, img.size[0] - quiet_zone), fill=255
    )

    draw.rectangle((58, 58, 92, 92), fill=0)
    draw.rectangle((img.size[0] - 58, 92, img.size[0] - 92, 58), fill=0)
    draw.rectangle((55, img.size[0] - 92, 92, img.size[0] - 58), fill=0)

    round_eyes = Image.new("L", img.size, 255)

    draw = ImageDraw.Draw(round_eyes)
    draw.ellipse((quiet_zone, quiet_zone, eye_border, eye_border), fill=None, outline=0, width=8)
    draw.ellipse(
        (img.size[0] - eye_border, quiet_zone, img.size[0] - quiet_zone, eye_border),
        fill=None,
        outline=0,
        width=8,
    )
    draw.ellipse(
        (quiet_zone, img.size[0] - eye_border, eye_border, img.size[0] - quiet_zone),
        fill=None,
        outline=0,
        width=8,
    )

    final = Image.composite(round_eyes, img, eyes_mask)

    return final


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """

    # fmt: off
    stimuli = {
        "logo": logo(),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



"""Tests whether resolving of resolution/components works as expected

Separately tests each of three components: visual size, shape, ppd.
For each component, the resolver routine has 1 expected behavior:
from the two givens, calculate the third unknown component.
As long as the two givens are valid (see valid_*),
the output will be the third component in canonical form.

Also tests the overall resolving: given three inputs with one None, resolve the unknown.
This function has two expected behaviors:
- if the resolution can be resolved, i.e., there is just 1 unknown per dimension,
returns all three components in canonical form
- if the resolution cannot be resolved, i.e., there is more than 1 unknown in a dimension,
raises a ValueError.

The tests do this through parameterization:
pytest runs each test for each of the specified parameter sets.

For the valid-input tests, each parameter set specifies the (valid) input,
and value (in canonical form) that should be returned.
If the actual returned value matches this expected return value, the test passes.

For the invalid-input tests, each parameter set specifies the (invalid) input,
and the Exception that should be raised.
If the function under testing raises this specified exception, the test passes.
"""
import pytest

from stimupy.utils import resolution


#############################
#   ppd from shape & size   #
#############################
@pytest.mark.parametrize(
    "shape, visual_size, ppd",
    [
        ((1024, 1024), (32, 32), (32, 32)),
        ((1024, 1024), (16, 16), (64, 64)),
        ((512, 512), (32, 32), (16, 16)),
        (1024, (32, 32), (32, 32)),
        ((1024, 1024), 32, (32, 32)),
        ([1024, 768], (32, 48), (32, 16)),
        ([None, None], (32, 24), (None, None)),
        ([512, None], (32, 24), (16, None)),
    ],
)
def test_ppd_from_shape_visual_size(shape, visual_size, ppd):
    out = resolution.ppd_from_shape_visual_size(shape, visual_size)
    assert out.horizontal == ppd[1] and out.vertical == ppd[0]


#############################
#   shape from size & ppd   #
#############################
@pytest.mark.parametrize(
    "shape, visual_size, ppd",
    [
        ((1024, 1024), (32, 32), (32, 32)),
        ((1024, 1024), (16, 16), (64, 64)),
        ((512, 512), (32, 32), (16, 16)),
        #    (1024, (32, 32), (32, 32)),
        ((1024, 1024), 32, (32, 32)),
        ([1024, 768], (32, 48), (32, 16)),
        ([None, None], (32, 24), (None, None)),
        ([512, None], (32, 24), (16, None)),
    ],
)
def test_shape_from_visual_size_ppd(visual_size, ppd, shape):
    out = resolution.shape_from_visual_size_ppd(visual_size, ppd)
    assert out.height == shape[0] and out.width == shape[1]


#############################
#   size from shape & ppd   #
#############################
@pytest.mark.parametrize(
    "shape, visual_size, ppd",
    [
        ((1024, 1024), (32, 32), (32, 32)),
        ((1024, 1024), (16, 16), (64, 64)),
        ((512, 512), (32, 32), (16, 16)),
        (1024, (32, 32), (32, 32)),
        ((1024, 1024), (32, 32), 32),
        ([1024, 768], (32, 48), (32, 16)),
        ([None, None], (None, None), (32, 24)),
        ([512, None], (32, None), (16, 24)),
    ],
)
def test_visual_size_from_shape_ppd(visual_size, ppd, shape):
    out = resolution.visual_size_from_shape_ppd(shape=shape, ppd=ppd)
    assert out.height == visual_size[0] and out.width == visual_size[1]


#############################
#    resolve resolution     #
#############################
@pytest.mark.parametrize(
    "shape, visual_size, ppd",
    [
        ((1024, 1024), (32, 32), (32, 32)),  # Already resolved
        (None, (32, 32), (32, 32)),  # One unknown
        ((1024, 1024), None, (32, 32)),  # One unknown
        ((1024, 1024), (32, 32), None),  # One unknown
        ((None, 1024), (32, None), (32, 32)),  # One unknown in each dimension
        ((None, 1024), (32, 32), (32, None)),  # One unknown in each dimension
        ((1024, None), (32, 32), (None, 32)),  # One unknown in each dimension
        (1024, 32, None),
    ],
)
def test_resolve_resolution(shape, visual_size, ppd):
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    resolution.valid_resolution(shape=shape, visual_size=visual_size, ppd=ppd)


@pytest.mark.parametrize(
    "shape, visual_size, ppd",
    [
        (None, None, (32, 32)),  # Two unknowns
        ((1024, 1024), None, None),  # Two unknowns
        (None, (32, 32), None),  # Two unknowns
        ((1024, None), (32, None), (32, 32)),  # Two unknowns in one dimension
        ((1024, 1024), (32, None), (32, None)),  # Two unknowns in one dimension
        ((1024, None), (32, 32), (32, None)),  # Two unknowns in one dimension
    ],
)
def test_too_many_unknowns(shape, visual_size, ppd):
    with pytest.raises(ValueError):
        resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)



"""Stimuli from Murray (2020) https://doi.org/10/gh57gf

This module reproduces most of the stimuli used by Murray (2020)
as they were provided to the model described in that paper but
normalized between 0 and 1.
The stimuli are show in Fig 1 of the paper.
NOTE that the Haze illusion (Fig 1m) is not provided.

Each stimulus is provided by a separate function,
a full list can be found as stimupy.papers.murray2020.__all__

The output of each of these functions is a stimulus dictionary.

For a visual representation of all the stimuli and their mask,
simply run this module as a script:

    $ python stimuli/papers/murray2020.py

Attributes
----------
__all__ (list of str): list of all stimulus-functions
    that are exported by this module when executing
        >>> from stimupy.papers.murray2020 import *

References
-----------
Murray, R. F. (2020).
    A model of lightness perception
    guided by probabilistic assumptions about lighting and reflectance.
    Journal of Vision, 20(7), 28.
    https://doi.org/10/gh57gf
"""

import os.path

import numpy as np
import scipy.io

import stimupy
from stimupy.utils import pad_dict_by_visual_size, rotate_dict

__all__ = [
    "argyle",
    "argyle_control",
    "argyle_long",
    "snake",
    "snake_control",
    "koffka_adelson",
    "koffka_broken",
    "koffka_connected",
    "checkassim",
    "simcon",
    "simcon_articulated",
    "white",
]

data_dir = os.path.dirname(__file__)
mat_fname = os.path.join(data_dir, "murray2020.mat")
mat_content = scipy.io.loadmat(mat_fname)

PPD = 16 / 8.0
PAD = True


def gen_all(skip=False):
    stims = {}  # save the stimulus-dicts in a larger dict, with name as key
    for stim_name in __all__:
        print(f"Generating murray2020.{stim_name}")

        # Get a reference to the actual function
        func = globals()[stim_name]
        try:
            stim = func()

            # Accumulate
            stims[stim_name] = stim
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stims


def get_mask(target1, target2, shape):
    mask = np.zeros(shape)

    # Convert MATLAB 1-based index to numpy 0-based index
    target1 = target1 - 1
    target2 = target2 - 1

    # Fill target1 mask
    for x in range(target1[1], target1[3] + 1):
        for y in range(target1[0], target1[2] + 1):
            mask[y][x] = 1

    # Fill target2 mask
    for x in range(target2[1], target2[3] + 1):
        for y in range(target2[0], target2[2] + 1):
            mask[y, x] = 2

    mask = np.array(mask, dtype=int)

    return mask


def argyle(ppd=PPD):
    """Argyle illusion, Murray (2020) Fig 1a

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["argyle"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.4875,
        "CI95_proportion_expected": (-0.154903, 0.154903),
    }

    params = {
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return {"img": normed_img, "target_mask": mask, **params}


def argyle_control(ppd=PPD):
    """Argyle control figure, Murray (2020) Fig 1c

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["argyle_control"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.35,
        "CI95_proportion_expected": (-0.147814, 0.147814),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def argyle_long(ppd=PPD):
    """Long-range Argyle illusion, Murray (2020) Fig 1b

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["argyle_long"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.6,
        "CI95_proportion_expected": (-0.151821, 0.151821),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def snake(ppd=PPD):
    """Snake illusion, Murray (2020) Fig 1i

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["snake"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.9625,
        "CI95_proportion_expected": (-0.0588765, 0.0588765),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def snake_control(ppd=PPD):
    """Snake control figure, Murray (2020) Fig 1j

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["snake_control"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.8,
        "CI95_proportion_expected": (-0.123961, 0.123961),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def koffka_adelson(ppd=PPD):
    """Koffka-Adelson figure, Murray (2020) Fig 1e

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["koffka_adelson"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.8375,
        "CI95_proportion_expected": (-0.114326, 0.114326),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def koffka_broken(ppd=PPD):
    """Koffka ring, broken, Murray (2020) Fig 1d

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["koffka_broken"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.8,
        "CI95_proportion_expected": (-0.123961, 0.123961),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def koffka_connected(ppd=PPD):
    """Koffka ring, connected, Murray (2020) Fig 1f

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["koffka_connected"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.6,
        "CI95_proportion_expected": (-0.151821, 0.151821),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def checkassim(ppd=PPD, pad=PAD):
    """Checkerboard assimilation, Murray (2020) Fig 1h

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    params = {
        "ppd": ppd,
        "board_shape": (7, 10),
        "check_visual_size": 1 / PPD,
        "target_indices": ((3, 6), (3, 3)),
        "extend_targets": False,
        "intensity_checks": (17.5, 70.0),
        "intensity_target": 35.0,
    }
    stim = stimupy.checkerboards.checkerboard(
        **params,
    )

    if pad:
        padding = np.array(((4, 5), (3, 3))) / PPD
        stim = pad_dict_by_visual_size(stim, padding, ppd=ppd, pad_value=35.0)
        params["padding"] = padding

    # Normalize intensity values to [0, 1]
    original_range = (stim["img"].min(), stim["img"].max())
    stim["img"] = (stim["img"] - stim["img"].min()) / (stim["img"].max() - stim["img"].min())

    experimental_data = {
        "mean_proportion_expected": 0.8625,
        "CI95_proportion_expected": (-0.106723, 0.106723),
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        original_range=original_range,
        intensity_range=(0.0, 1.0),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def simcon(ppd=PPD):
    """Classic simultaneous contrast figure, Murray (2020) Fig 1k

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["simcon"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.925,
        "CI95_proportion_expected": (-0.0816258, 0.0816258),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def simcon_articulated(ppd=PPD):
    """Articulated simultaneous contrast figure, Murray (2020) Fig 1l

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    a = mat_content["simcon_articulated"]
    img = np.array(((a[0])[0])[0])
    target1 = np.array((((a[0])[0])[1])[0])
    target2 = np.array((((a[0])[0])[2])[0])
    mask = get_mask(target1, target2, img.shape)

    img = img.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)
    mask = mask.repeat(repeats=int(ppd / PPD), axis=0).repeat(repeats=int(ppd / PPD), axis=1)

    # Normalize intensity values to [0, 1]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 1.0,
        "CI95_proportion_expected": (-0.0, 0.0),
    }

    stim = {
        "img": normed_img,
        "target_mask": mask.astype(int),
        "visual_size": np.array(normed_img.shape) / ppd,
        "shape": normed_img.shape,
        "ppd": ppd,
        "intensity_range": (0.0, 1.0),
        "original_range": original_range,
        "experimental_data": experimental_data,
    }
    return stim


def white(ppd=PPD):
    """White's illusion, Murray (2020) Fig 1A

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Murray, R. F. (2020).
        A model of lightness perception
        guided by probabilistic assumptions about lighting and reflectance.
        Journal of Vision, 20(7), 28.
        https://doi.org/10/gh57gf
    """

    params = {
        "ppd": ppd,
        "visual_size": (8.0, 8.0),
        "frequency": 4.0 / 8.0,
        "target_indices_top": (2, 4, 6),
        "target_indices_bottom": (3, 5, 7),
        "target_center_offset": 2,
        "target_heights": 2,
        "intensity_bars": (17.5, 70),
        "intensity_target": 35.0,
        "period": "even",
    }

    stim = stimupy.whites.white_two_rows(**params)
    stim = rotate_dict(stim)
    reduced_mask = np.where(stim["target_mask"] == 2, 2, 0)
    reduced_mask = np.where(stim["target_mask"] == 5, 1, reduced_mask).astype(int)

    # Normalize intensity values to [0, 1]
    img = stim["img"]
    original_range = (img.min(), img.max())
    normed_img = (img - img.min()) / (img.max() - img.min())

    experimental_data = {
        "mean_proportion_expected": 0.95,
        "CI95_proportion_expected": (-0.0675418, 0.0675418),
    }

    params.update(
        visual_size=np.array(normed_img.shape) / ppd,
        shape=normed_img.shape,
        original_range=original_range,
        intensity_range=(0.0, 1.0),
        experimental_data=experimental_data,
    )
    return {"img": normed_img, "target_mask": reduced_mask, **params}


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    # Generate all stimuli exported in __all__
    stims = gen_all(skip=True)
    plot_stimuli(stims, mask=True)



import copy
import warnings
from collections import namedtuple

import numpy as np

__all__ = [
    "resolve",
    "resolve_1D",
    "resolve_dict",
    "visual_angle_from_length_ppd",
    "visual_angles_from_lengths_ppd",
    "visual_size_from_shape_ppd",
    "length_from_visual_angle_ppd",
    "lengths_from_visual_angles_ppd",
    "shape_from_visual_size_ppd",
    "ppd_from_shape_visual_size",
    "ppd_from_length_visual_angle",
    "compute_ppd",
    "validate_shape",
    "validate_ppd",
    "validate_visual_size",
    "valid_1D",
    "valid_resolution",
    "valid_dict",
]

Visual_size = namedtuple("Visual_size", "height width")
Shape = namedtuple("Shape", "height width")
Ppd = namedtuple("Ppd", "vertical horizontal")


class ResolutionError(ValueError):
    pass


class TooManyUnknownsError(ValueError):
    pass


def resolve(shape=None, visual_size=None, ppd=None):
    """Resolves the full resolution, for 2 givens and 1 unknown

    A resolution consists of a visual size in degrees, a shape in pixels,
    and specification of the number of pixels per degree.
    Since there is a strict geometric relation between these,
    shape = visual_size * ppd,
    if two are given, the third can be calculated using this function.

    This function resolves the resolution in both dimensions.


    Parameters
    ----------
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] in pixels
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]

    Returns
    -------
    Shape NamedTuple, with two attributes:
        .height: int, height in pixels
        .width: int, width in pixels
        See validate_shape
    Visual_size NamedTuple, with two attributes:
        .height: float, height in degrees visual angle
        .width: float, width in degrees visual angle
        See validate_visual_size
    ppd NamedTuple, with two attributes:
        .vertical: int, vertical pixels per degree (ppd)
        .horizontal: int, horizontal pixels per degree (ppd)
        see validate_ppd
    """

    # Canonize inputs
    ppd = validate_ppd(ppd)
    visual_size = validate_visual_size(visual_size)
    shape = validate_shape(shape)

    # Vertical
    height, visual_height, ppd_vertical = resolve_1D(
        length=shape.height, visual_angle=visual_size.height, ppd=ppd.vertical
    )

    # Horizontal
    width, visual_width, ppd_horizontal = resolve_1D(
        length=shape.width, visual_angle=visual_size.width, ppd=ppd.horizontal
    )

    # Check & canonize outputs
    shape = validate_shape((height, width))
    ppd = validate_ppd((ppd_vertical, ppd_horizontal))
    visual_size = validate_visual_size((visual_height, visual_width))

    # Assert that resolved resolution is valid
    valid_resolution(shape=shape, visual_size=visual_size, ppd=ppd)

    return shape, visual_size, ppd


def resolve_1D(length=None, visual_angle=None, ppd=None, round=True):
    """Resolves the full resolution, for 2 givens and 1 unknown

    A resolution consists of a visual size in degrees, a shape in pixels,
    and specification of the number of pixels per degree.
    Since there is a strict geometric relation between these,
    shape = visual_size * ppd,
    if two are given, the third can be calculated using this function.

    This function resolves the resolution in a single dimension.


    Parameters
    ----------
    length : Number, length in pixels, or None (default)
    visual_angle : Number, length in degrees, or None (default)
    ppd : Number, pixels per degree, or None (default)

    Returns
    -------
    length : int, length in pixels
    visual_angle : float, length in degrees
    ppd : float, pixels per degree
    """

    # How many unknowns passed in?
    n_unknowns = (length is None) + (visual_angle is None) + (ppd is None)

    # Triage based on number of unknowns
    if n_unknowns > 1:  # More than 1 unknown we cannot resolve
        raise TooManyUnknownsError(
            f"Too many unknowns to resolve resolution; {length},{visual_angle},{ppd}"
        )
    else:  # 1 unknown, so need to resolve
        # Which unknown?
        if length is None:
            if round:
                visual_angle_old = np.round(copy.deepcopy(visual_angle), 10)
                visual_angle = np.floor(np.round(visual_angle * ppd, 10)) / ppd
                if visual_angle_old != visual_angle:
                    warnings.warn(
                        f"Rounding visual angle because of ppd; {visual_angle_old} ->"
                        f" {visual_angle}"
                    )
            length = length_from_visual_angle_ppd(visual_angle=visual_angle, ppd=ppd, round=round)
        elif visual_angle is None:
            visual_angle = visual_angle_from_length_ppd(length=length, ppd=ppd)
        elif ppd is None:
            ppd = ppd_from_length_visual_angle(length=length, visual_angle=visual_angle)

    return length, visual_angle, ppd


def resolve_dict(dct):
    """Resolves the full resolution ("shape", "ppd", "visual_size"), for 2
    givens and 1 unknown in the input dictionary

    A resolution consists of a visual size in degrees, a shape in pixels,
    and specification of the number of pixels per degree.
    Since there is a strict geometric relation between these,
    shape = visual_size * ppd,
    if two are given, the third can be calculated using this function.

    This function resolves the resolution in both dimensions.

    Parameters
    ----------
    dct : dict
        dictionary with at least two out the three keys: "shape", "ppd", "visual_size"

    Returns
    -------
    Resolved dict

    See also
    ---------
    stimupy.utils.resolution
    """

    # Resolve
    ppd = dct["ppd"] if "ppd" in dct.keys() else None
    shape = dct["shape"] if "shape" in dct.keys() else None
    visual_size = dct["visual_size"] if "visual_size" in dct.keys() else None
    shape, visual_size, ppd = resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    # Update dict
    dct["shape"] = shape
    dct["visual_size"] = visual_size
    dct["ppd"] = ppd
    return


def visual_size_to_axes(visual_size, shape, origin="mean"):
    """Generate axes from visual size, shape and origin

    Parameters
    ----------
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] in pixels
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] in degrees
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    (x, y):
        x and y axes, linearly spacing visual size in intervals defined by shape,
        and with the origin set in the right place.
    """

    # Validate params
    visual_size = validate_visual_size(visual_size=visual_size)
    shape = validate_shape(shape)

    # Set origin
    if origin == "corner":
        x = np.linspace(0, visual_size.width, shape.width)
        y = np.linspace(0, visual_size.height, shape.height)
    elif origin == "mean":
        vrange = (visual_size.height / 2, visual_size.width / 2)
        x = np.linspace(-vrange[1], vrange[1], shape.width)
        y = np.linspace(-vrange[0], vrange[0], shape.height)
    elif origin == "center":
        vrange = (visual_size.height / 2, visual_size.width / 2)
        x = np.linspace(-vrange[1], vrange[1], shape.width, endpoint=False)
        y = np.linspace(-vrange[0], vrange[0], shape.height, endpoint=False)
    else:
        raise ValueError('origin can only be be "corner", "mean" or "center"')

    return x, y


#############################
#    Resolve components     #
#############################
def visual_angle_from_length_ppd(length, ppd):
    """Calculate visual angle (degrees) from length (pixels) and pixels-per-degree

    Parameters
    ----------
    length : int or None
        length in pixels
    ppd : int or None
        pixels per degree

    Returns
    -------
    Length (pixels) translated to visual angle (degrees)
    """
    if length is not None and ppd is not None:
        visual_angle = length / ppd
    else:
        visual_angle = None
    return visual_angle


def visual_angles_from_lengths_ppd(lengths, ppd):
    """Calculate visual sizes (degrees) from given shapes (pixels) and pixels-per-degree

    Parameters
    ----------
    lengths : Sequence[int, int, ...] or None
        list of lengths
    ppd : int or None
        pixels per degree

    Returns
    -------
    List with lengths (pixels) translated to visual angles (degrees)
    """
    if isinstance(lengths, (int, float)):
        lengths = [
            lengths,
        ]

    if lengths is not None and ppd is not None:
        visual_angles = []
        for length in lengths:
            visual_angles.append(visual_angle_from_length_ppd(length, ppd))
    else:
        visual_angles = None

    if len(visual_angles) == 1:
        visual_angles = visual_angles[0]
    return visual_angles


def visual_size_from_shape_ppd(shape, ppd):
    """Calculate visual size (degrees) from given shape (pixels) and pixels-per-degree

    Parameters
    ----------
    shape : Sequence[int, int]; or int, or None
        each element has to be of type that can be cast to int, or None.
        See validate_shape
    ppd : Sequence[int, int]; or int, or None
        each element has to be of type that can be cast to int, or None.
        See validate_ppd

    Returns
    -------
    Visual_size NamedTuple, with two attributes:
        .height: float, height in degrees visual angle
        .width: float, width in degrees visual angle
        See validate_visual_size
    """

    # Canonize inputs
    shape = validate_shape(shape)
    ppd = validate_ppd(ppd)

    # Calculate width and height in pixels
    width = visual_angle_from_length_ppd(shape.width, ppd.horizontal)
    height = visual_angle_from_length_ppd(shape.height, ppd.vertical)

    # Construct Visual size NamedTuple:
    visual_size = Visual_size(width=width, height=height)

    return visual_size


def length_from_visual_angle_ppd(visual_angle, ppd, round=True):
    """Calculate length (pixels) from visual angle (degrees) and pixels-per-degree

    Parameters
    ----------
    visual_angle : float or None
        visual angle in degrees
    ppd : int or None
        pixels per degree
    round : bool
        if True, round output length to full pixels

    Returns
    -------
    visual angle (degrees) translated to length (pixels)
    """
    if visual_angle is not None and ppd is not None:
        fpix = np.round(visual_angle * ppd, 10)

        if round:
            pix = int(fpix)
            if fpix > 0 and fpix % pix:
                warnings.warn(f"Rounding shape; {visual_angle} * {ppd} = {fpix} -> {pix}")
        else:
            pix = float(fpix)
    else:
        pix = None
    return pix


def lengths_from_visual_angles_ppd(visual_angles, ppd, round=True):
    """Calculate lengths (pixels) from visual angles (degrees) and pixels-per-degree

    Parameters
    ----------
    visual_angles : Sequence[float, float, ...] or None
        list of visual angles
    ppd : int or None
        pixels per degree
    round : bool
        if True, round output length to full pixels

    Returns
    -------
    List with visual angles (degrees) translated to lengths (pixels)
    """
    if isinstance(visual_angles, (int, float, np.int64)):
        visual_angles = [
            visual_angles,
        ]
    if isinstance(ppd, (list, tuple)):
        raise ValueError("ppd should be a single number")

    if visual_angles is not None and ppd is not None:
        lengths = []
        for angle in visual_angles:
            lengths.append(length_from_visual_angle_ppd(angle, ppd, round=round))
    else:
        lengths = None

    if len(lengths) == 1:
        lengths = lengths[0]
    return lengths


def shape_from_visual_size_ppd(visual_size, ppd):
    """Calculate shape (pixels) from given visual size (degrees) and pixels-per-degree

    Parameters
    ----------
    visual_size : Sequence[Number, Number]; or Number; or None
        each element has to be of type that can be cast to float, or None.
    ppd : Sequence[int, int]; or int, or None
        each element has to be of type that can be cast to int, or None.
        See validate_ppd

    Returns
    -------
    Shape NamedTuple, with two attributes:
        .height: int, height in pixels
        .width: int, width in pixels
        See validate_shape
    """

    # Canonize inputs
    visual_size = validate_visual_size(visual_size)
    ppd = validate_ppd(ppd)

    # Calculate width and height in pixels
    width = length_from_visual_angle_ppd(visual_size.width, ppd.horizontal)
    height = length_from_visual_angle_ppd(visual_size.height, ppd.vertical)

    # Construct Shape NamedTuple:
    shape = Shape(width=width, height=height)

    return shape


def ppd_from_shape_visual_size(shape, visual_size):
    """Calculate resolution (ppd) from given shape (pixels) and visual size (degrees)

    Parameters
    ----------
    shape : Sequence[int, int]; or int, or None
        each element has to be of type that can be cast to int, or None.
        See validate_shape
    visual_size : Sequence[Number, Number]; or Number; or None
        each element has to be of type that can be cast to float, or None.
        See validate_visual_size

    Returns
    -------
    ppd NamedTuple, with two attributes:
        .vertical: int, vertical pixels per degree (ppd)
        .horizontal: int, horizontal pixels per degree (ppd)
        see validate_ppd
    """

    # Canonize inputs
    shape = validate_shape(shape)
    visual_size = validate_visual_size(visual_size)

    # Calculate horizontal and vertical ppds
    horizontal = ppd_from_length_visual_angle(shape.width, visual_size.width)
    vertical = ppd_from_length_visual_angle(shape.height, visual_size.height)

    # Construct Ppd NamedTuple
    ppd = Ppd(horizontal=horizontal, vertical=vertical)

    return ppd


def ppd_from_length_visual_angle(length, visual_angle):
    """Calculate pixels-per-degree from length (pixels) and visual angle (degrees)

    Parameters
    ----------
    length : int or None
        length in pixels
    visual_angle : float or None
        visual angle in degrees

    Returns
    -------
    visual angle (degrees) translated to length (pixels)
    """
    if visual_angle is not None and length is not None:
        ppd = length / visual_angle
    else:
        ppd = None
    return ppd


def compute_ppd(screen_size, resolution, distance):
    """Compute the pixels per degree in a presentation setup
    i.e., the number of pixels in the central one degree of visual angle

    Parameters
    ----------
    screen_size : (float, float)
        physical size, in whatever units you prefer, of the presentation screen
    resolution : (float, float)
        screen resolution, in pixels,
        in the same direction that screen size was measured in
    distance : float
        physical distance between the observer and the screen, in the same unit as screen_size

    Returns
    -------
    float
        ppd, the number of pixels in one degree of visual angle
    """

    ppmm = resolution / screen_size
    mmpd = 2 * np.tan(np.radians(0.5)) * distance
    return ppmm * mmpd


#############################
#    Validate components    #
#############################
def validate_shape(shape):
    """Put specification of shape (in pixels) in canonical form, if possible

    Parameters
    ----------
    shape : Sequence of length 1 or 2; or None
        if 2 elements: interpret as (height, width)
        if 1 element: use as both height and width
        if None: return (None, None)
        each element has to be of type that can be cast to int, or None.

    Returns
    -------
    Shape NamedTuple, with two attributes:
        .height: int, height in pixels
        .width: int, width in pixels

    Raises
    ------
    ValueError
        if input does not have at least 1 element
    TypeError
        if input is not a Sequence(int, int) and cannot be cast to one
    ValueError
        if input has more than 2 elements
    """

    # Check if string:
    if isinstance(shape, str):
        shape = float(shape)

    # Check if sequence
    try:
        if len(shape) < 1:
            # Empty sequence
            raise TypeError(f"shape must be of at least length 1: {shape}")
    except TypeError:  # not a sequence; make it one
        shape = (shape, shape)

    # Check if length == 2
    if len(shape) == 1:
        # If Sequence of len()=1 is passed in, use as both height and width
        shape = (shape[0], shape[0])
    elif len(shape) > 2:
        # If Sequence of len()>2 is passed in: error
        raise TypeError(f"shape must be of length 1 or 2, not greater: {shape}")

    # Unpack
    width = shape[1]
    height = shape[0]

    # TODO: check if whole integer?

    # Convert to int
    if width is not None:
        width = int(width)
    if height is not None:
        height = int(height)

    # Check non-negative
    if (width is not None and width <= 0) or (height is not None and height <= 0):
        raise ValueError(f"shape has to be positive; {width, height}")

    # Initiate namedtuple:
    return Shape(height=height, width=width)


def validate_ppd(ppd):
    """Put specification of ppd in canonical form, if possible

    Parameters
    ----------
    ppd : Sequence of length 1 or 2; or None
        if 2 elements: interpret as (vertical, horizontal)
        if 1 element: use as both vertical and horizontal
        if None: return (None, None)
        each element has to be of type that can be cast to int, or None.

    Returns
    -------
    ppd NamedTuple, with two attributes:
        .vertical: int, vertical pixels per degree (ppd)
        .horizontal: int, horizontal pixels per degree (ppd)

    Raises
    ------
    ValueError
        if input does not have at least 1 element
    TypeError
        if input is not a Sequence(int, int) and cannot be cast to one
    ValueError
        if input has more than 2 elements
    """

    # Check if string:
    if isinstance(ppd, str):
        ppd = float(ppd)

    # Check if sequence
    try:
        if len(ppd) < 1:
            # Empty sequence
            raise TypeError(f"ppd must be of at least length 1: {ppd}")
    except TypeError:  # not a sequence; make it one
        ppd = (ppd, ppd)

    # Check if length == 2
    if len(ppd) == 1:
        # If Sequence of len()=1 is passed in, use as both height and width
        ppd = (ppd[0], ppd[0])
    elif len(ppd) > 2:
        # If Sequence of len()>2 is passed in: error
        raise TypeError(f"ppd must be of length 1 or 2, not greater: {ppd}")

    # Unpack
    horizontal = ppd[1]
    vertical = ppd[0]

    # Convert to float
    if horizontal is not None:
        horizontal = float(horizontal)
    if vertical is not None:
        vertical = float(vertical)

    # Check non-negative
    if (horizontal is not None and horizontal <= 0) or (vertical is not None and vertical <= 0):
        raise ValueError(f"ppd has to be positive; {horizontal, vertical}")

    # Initiate namedtuple:
    return Ppd(horizontal=horizontal, vertical=vertical)


def validate_visual_size(visual_size):
    """Put specification of visual size in canonical form, if possible

    Parameters
    ----------
    visual_size : Sequence of length 1 or 2; or None
        if 2 elements: interpret as (height, width)
        if 1 element: use as both height and width
        if None: return (None, None)
        each element has to be of type that can be cast to float, or None.

    Returns
    -------
    Visual_size NamedTuple, with two attributes:
        .height: float, height in degrees visual angle
        .width: float, width in degrees visual angle

    Raises
    ------
    ValueError
        if input does not have at least 1 element
    TypeError
        if input is not a Sequence(float, float) and cannot be cast to one
    ValueError
        if input has more than 2 elements
    """

    # Check if string:
    if isinstance(visual_size, str):
        visual_size = float(visual_size)

    # Check if sequence
    try:
        if len(visual_size) < 1:
            # Empty sequence
            raise TypeError(f"visual_size must be of at least length 1: {visual_size}")
    except TypeError:  # not a sequence; make it one
        visual_size = (visual_size, visual_size)

    # Check if length == 2
    if len(visual_size) == 1:
        # If Sequence of len()=1 is passed in, use as both height and width
        visual_size = (visual_size[0], visual_size[0])
    elif len(visual_size) > 2:
        # If Sequence of len()>2 is passed in: error
        raise TypeError(f"visual_size must be of length 1 or 2, not greater: {visual_size}")

    # Unpack
    width = visual_size[1]
    height = visual_size[0]

    # Convert to float
    if width is not None:
        width = float(width)
    if height is not None:
        height = float(height)

    # Check non-negative
    if (width is not None and width < 0) or (height is not None and height < 0):
        raise ValueError(f"visual_size has to be nonnegative; {width, height}")

    # Initiate namedtuple:
    return Visual_size(height=height, width=width)


def valid_1D(length, visual_angle, ppd):
    """Asserts that the combined specification of resolution is geometrically valid.

    Asserts the combined specification of shape (in pixels), visual_size (deg) and ppd.
    If this makes sense, i.e. (roughly), int(visual_size * ppd) == shape,
    this function passes without output.
    If the specification does not make sense, raises a ResolutionError.

    Note that the resolution specification has to be fully resolved,
    i.e., none of the parameters can be None

    Parameters
    ----------
    length : int, length in pixels
    visual_angle : float, size in degrees
    ppd : int, resolution in pixels-per-degree

    Raises
    ------
    ResolutionError
        if resolution specification is invalid,
        i.e. (roughly), if int(visual_angle * ppd) != length
    """

    # Check by calculating one component
    calculated = length_from_visual_angle_ppd(visual_angle=visual_angle, ppd=ppd)
    if calculated != length:
        raise ResolutionError(f"Invalid resolution; {visual_angle},{length},{ppd}")


def valid_resolution(shape, visual_size, ppd):
    """Asserts that the combined specification of resolution is geometrically valid.

    Asserts the combined specification of shape (in pixels), visual_size (deg) and ppd.
    If this makes sense, i.e. (roughly), int(visual_size * ppd) == shape,
    this function passes without output.
    If the specification does not make sense, raises a ResolutionError.

    Note that the resolution specification has to be fully resolved,
    i.e., none of the parameters can be/contain None

    Parameters
    ----------
    shape : 2-tuple (height, width), or something that can be cast (see validate_shape)
    visual_size : 2-tuple (height, width), or something that can be cast (see validate_visual_size)
    ppd : 2-tuple (vertical, horizontal), or something that can be cast (see validate_ppd)

    Raises
    ------
    ResolutionError
        if resolution specification is invalid,
        i.e. (roughly), if int(visual_size * ppd) != shape
    """

    # Canonize inputs
    shape = validate_shape(shape)
    ppd = validate_ppd(ppd)
    visual_size = validate_visual_size(visual_size)

    # Check by calculating one component
    calculated = shape_from_visual_size_ppd(visual_size=visual_size, ppd=ppd)
    if calculated != shape:
        raise ResolutionError(f"Invalid resolution; {visual_size},{shape},{ppd}")


def valid_dict(dct):
    """Asserts that the combined specification of resolution in dict is geometrically valid.

    Asserts the combined specification of shape (in pixels), visual_size (deg) and ppd.
    If this makes sense, i.e. (roughly), int(visual_size * ppd) == shape,
    this function passes without output.
    If the specification does not make sense, raises a ResolutionError.

    Note that the resolution specification has to be fully resolved,
    i.e., none of the parameters can be/contain None

    Parameters
    ----------
    dct : dict
        dictionary with at least the keys "shape", "ppd", "visual_size"

    Raises
    ------
    ResolutionError
        if resolution specification is invalid,
        i.e. (roughly), if int(visual_size * ppd) != shape
    """
    ppd = dct["ppd"] if "ppd" in dct.keys() else None
    shape = dct["shape"] if "shape" in dct.keys() else None
    visual_size = dct["visual_size"] if "visual_size" in dct.keys() else None

    # Assert that resolution is valid
    valid_resolution(shape=shape, visual_size=visual_size, ppd=ppd)



import numpy as np

from stimupy.noises import pseudo_white_spectrum
from stimupy.utils import resolution
from stimupy.utils.contrast_conversions import adapt_intensity_range

__all__ = [
    "one_over_f",
    "pink",
    "brown",
]


def one_over_f(
    visual_size=None,
    ppd=None,
    shape=None,
    exponent=None,
    intensity_range=(0, 1),
    pseudo_noise=False,
):
    """Draw 1 / (f**exponent) noise texture

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    exponent
        exponent used to create 1 / (f**exponent) noise
    intensity_range : Sequence[Number, Number]
        minimum and maximum intensity value; default: (0, 1)
    pseudo_noise : bool
        if True, generate pseudo-random noise with ideal power spectrum.

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        and additional keys containing stimulus parameters
    """
    if exponent is None:
        raise ValueError("one_over_f() missing argument 'exponent' which is not 'None'")
    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    # Prepare spatial frequency axes and create bandpass filter:
    fy = np.fft.fftshift(np.fft.fftfreq(shape[0], d=1.0 / np.unique(ppd)))
    fx = np.fft.fftshift(np.fft.fftfreq(shape[1], d=1.0 / np.unique(ppd)))
    Fx, Fy = np.meshgrid(fx, fy)

    # Create 2d array with 1 / (f**exponent)
    f = np.sqrt(Fy**2.0 + Fx**2.0)
    f = f**exponent
    f[f == 0.0] = 1.0  # Prevent division by zero (DC is zero anyways)

    if pseudo_noise:
        # Create white noise with frequency amplitude of 1 everywhere
        white_noise_fft = pseudo_white_spectrum(shape)
    else:
        # Create white noise and fft
        white_noise = np.random.rand(*shape) * 2.0 - 1.0
        white_noise_fft = np.fft.fftshift(np.fft.fft2(white_noise))

    # Create 1/f noise:
    noise_fft = white_noise_fft / f

    # ifft
    noise = np.fft.ifft2(np.fft.ifftshift(noise_fft))
    noise = np.real(noise)

    # Adjust intensity range:
    noise = adapt_intensity_range(noise, intensity_range[0], intensity_range[1])

    stim = {
        "img": noise,
        "noise_mask": None,
        "visual_size": visual_size,
        "ppd": ppd,
        "shape": shape,
        "exponent": exponent,
        "pseudo_noise": pseudo_noise,
        "intensity_range": [noise.min(), noise.max()],
    }
    return stim


def pink(
    visual_size=None,
    ppd=None,
    shape=None,
    intensity_range=(0, 1),
    pseudo_noise=False,
):
    """Draw pink (1 / f) noise texture

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    intensity_range : Sequence[Number, Number]
        minimum and maximum intensity value; default: (0, 1)
    pseudo_noise : bool
        if True, generate pseudo-random noise with ideal power spectrum

    Returns
    -------
    A stimulus dictionary with the noise array ['img']
    """

    stim = one_over_f(
        visual_size=visual_size,
        ppd=ppd,
        exponent=1.0,
        intensity_range=intensity_range,
        pseudo_noise=pseudo_noise,
    )
    return stim


def brown(
    visual_size=None,
    ppd=None,
    shape=None,
    intensity_range=(0, 1),
    pseudo_noise=False,
):
    """Draw brown (1 / (f**2.0)) noise texture

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    intensity_range : Sequence[Number, Number]
        minimum and maximum intensity value; default: (0, 1)
    pseudo_noise : bool
        if True, generate pseudo-random noise with ideal power spectrum

    Returns
    -------
    A stimulus dictionary with the noise array ['img']
    """

    stim = one_over_f(
        visual_size=visual_size,
        ppd=ppd,
        exponent=2.0,
        intensity_range=intensity_range,
        pseudo_noise=pseudo_noise,
    )
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 10,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "naturals_one-over-f": one_over_f(**default_params, exponent=0.5),
        "naturals_pink": pink(**default_params),
        "naturals_brown": brown(**default_params),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import numpy as np


def randomize_sign(array):
    """Randomize the sign of values in an array

    Parameters
    ----------
    array
        N-dimensional array

    Returns
    -------
    array
        Same array with randomized signs

    """
    sign = np.random.rand(*array.shape) - 0.5
    sign[sign <= 0.0] = -1.0
    sign[sign > 0.0] = 1.0
    array = array * sign
    return array


def pseudo_white_helper(
    shape,
    amplitude,
):
    """Generate pseudorandom white noise patch

    Parameters
    ----------
    shape : int or (int, int)
        Shape of noise patch in pixels (height, width)
    amplitude
        Amplitude of each (pos/neg) frequency component = A/2

    Returns
    -------
    output
        Pseudorandom white noise patch

    """
    if isinstance(shape, (float, int)):
        shape = (shape, shape)

    Re = np.random.rand(*shape) * amplitude - amplitude / 2.0
    Im = np.sqrt((amplitude / 2.0) ** 2 - Re**2)
    Im = randomize_sign(Im)
    output = Re + Im * 1j
    return output


def pseudo_white_spectrum(
    shape=(100, 100),
    amplitude=2.0,
):
    """Create pseudorandom white noise spectrum

    Code translated and adapted from Matlab scripts
    provided by T. Peromaa

    Parameters
    ----------
    shape : int or (int, int)
        Shape of noise array in pixels (height, width)
    amplitude
        Amplitude of noise power spectrum

    Returns
    -------
    spectrum
        Shifted 2d complex number spectrum. DC = 0.
        Amplitude of each (pos/neg) frequency component = A/2
        Power of each (pos/neg) frequency component = (A/2)**2

    """
    if isinstance(shape, (float, int)):
        shape = (shape, shape)

    y, x = shape
    A = amplitude

    if (y % 2 != 0) or (x % 2 != 0):
        raise ValueError("shape needs to be even-numbered")

    # We divide the noise spectrum in four quadrants with pseudorandom white noise
    quadrant1 = pseudo_white_helper((int(y / 2) - 1, int(x / 2) - 1), A)
    quadrant2 = pseudo_white_helper((int(y / 2) - 1, int(x / 2) - 1), A)
    quadrant3 = quadrant2[::-1, ::-1].conj()
    quadrant4 = quadrant1[::-1, ::-1].conj()

    # We place the quadrants in the spectrum to eventuate that each frequency component has
    # an amplitude of A/2
    spectrum = np.zeros([y, x], dtype=complex)
    spectrum[1 : int(y / 2), 1 : int(x / 2)] = quadrant1
    spectrum[1 : int(y / 2), int(x / 2) + 1 : x] = quadrant2
    spectrum[int(y / 2 + 1) : y, 1 : int(x / 2)] = quadrant3
    spectrum[int(y / 2 + 1) : y, int(x / 2 + 1) : x] = quadrant4

    # We need to fill the rows / columns that the quadrants do not cover
    # Fill first row:
    row = pseudo_white_helper((1, x), A)
    apu = np.fliplr(row)
    row[0, int(x / 2 + 1) : x] = apu[0, int(x / 2) : x - 1].conj()
    spectrum[0, :] = np.squeeze(row)

    # Fill central row:
    row = pseudo_white_helper((1, x), A)
    apu = np.fliplr(row)
    row[0, int(x / 2 + 1) : x] = apu[0, int(x / 2) : x - 1].conj()
    spectrum[int(y / 2), :] = np.squeeze(row)

    # Fill first column:
    col = pseudo_white_helper((y, 1), A)
    apu = np.flipud(col)
    col[int(y / 2 + 1) : y, 0] = apu[int(y / 2) : y - 1, 0].conj()
    spectrum[:, int(x / 2)] = np.squeeze(col)

    # Fill central column:
    col = pseudo_white_helper((y, 1), A)
    apu = np.flipud(col)
    col[int(y / 2 + 1) : y, 0] = apu[int(y / 2) : y - 1, 0].conj()
    spectrum[:, 0] = np.squeeze(col)

    # Set amplitude at filled-corners to A/2:
    spectrum[0, 0] = -A / 2 + 0j
    spectrum[0, int(x / 2)] = -A / 2 + 0j
    spectrum[int(y / 2), 0] = -A / 2 + 0j

    # Set DC = 0:
    spectrum[int(y / 2), int(x / 2)] = 0 + 0j
    return spectrum


from stimupy.noises import binaries, narrowbands, naturals, whites

__all__ = [
    "overview",
    "plot_overview",
    "pseudo_white_spectrum",
    "binaries",
    "narrowbands",
    "naturals",
    "whites",
]


def overview(skip=False):
    """Generate example stimuli from this module

    Returns
    -------
    dict[str, dict]
        Dict mapping names to individual stimulus dicts
    """
    stimuli = {}
    for stimmodule_name in __all__:
        if stimmodule_name in [
            "overview",
            "plot_overview",
            "pseudo_white_spectrum",
        ]:
            continue

        print(f"Generating stimuli from {stimmodule_name}")
        # Get a reference to the actual module
        # print(globals())
        stimmodule = globals()[stimmodule_name]

        print(stimmodule)
        try:
            stims = stimmodule.overview()

            # Accumulate
            stimuli.update(stims)
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stimuli


def plot_overview(mask=False, save=None, units="deg"):
    """Plot overview of examples in this module (and submodules)

    Parameters
    ----------
    mask : bool or str, optional
        If True, plot mask on top of stimulus image (default: False).
        If string is provided, plot this key from stimulus dictionary as mask
    save : None or str, optional
        If None (default), do not save the plot.
        If string is provided, save plot under this name.
    units : "px", "deg" (default), or str
        what units to put on the axes, by default degrees visual angle ("deg").
        If a str other than "deg"(/"degrees") or "px"(/"pix"/"pixels") is passed,
        it must be the key to a tuple in stim

    """
    from stimupy.utils import plot_stimuli

    stims = overview(skip=True)
    plot_stimuli(stims, mask=mask, units=units, save=save)


if __name__ == "__main__":
    plot_overview()



import numpy as np
from scipy.signal import fftconvolve

from stimupy.utils import resolution
from stimupy.utils.pad import add_padding, remove_padding

__all__ = [
    "convolve",
    "bandpass",
]


def convolve(
    arr1,
    arr2,
    mode="same",
    axes=None,
    padding=False,
):
    """
    Convolve two N-dimensional arrays using FFT

    Parameters
    ----------
    arr1 : numpy.ndarray
        Input array 1
    arr2 : numpy.ndarray
        Input array 2
    mode : str {"full", "valid", "same"}, optional
        String which indicates the size of the output. The default is "same".
    axes : int or None (default), optional
        Axes over which to convolve. The default is over all axes
    padding : Bool
        if True, pad array before convolving

    Returns
    -------
    out : numpy.ndarray
        Output array

    """
    c = int(arr1.shape[0] / 2)
    if padding:
        arr1 = add_padding(arr1, c, arr1.mean())
    out = fftconvolve(arr1, arr2, mode, axes)
    if padding:
        out = remove_padding(out, c)
    return out


def bandpass(
    visual_size=None,
    ppd=None,
    shape=None,
    center_frequency=None,
    bandwidth=None,
):
    """
    Function to create a 2d bandpass filter in the frequency domain

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    center_frequency : float
        center frequency of filter in cpd
    bandwidth : float
        bandwidth of filter in octaves

    Returns
    -------
    dict[str, Any]
        dict with the filter (key: "img"),
        and additional keys containing filter parameters
    """
    if center_frequency is None:
        raise ValueError("bandpass() missing argument 'center_frequency' which is not 'None'")
    if bandwidth is None:
        raise ValueError("bandpass() missing argument 'bandwidth' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    if center_frequency > (min(ppd) / 2):
        raise ValueError(
            f"Center frequency ({center_frequency}) should not exceed Nyquist limit {min(ppd)/2} (ppd/2)"
        )

    # Create frequency axes
    fy = np.fft.fftshift(np.fft.fftfreq(shape[0], d=1.0 / ppd[0]))
    fx = np.fft.fftshift(np.fft.fftfreq(shape[1], d=1.0 / ppd[1]))
    Fx, Fy = np.meshgrid(fx, fy)

    # Calculate the distance of each 2d spatial frequency from requested center frequency
    distance = np.abs(center_frequency - np.sqrt(Fx**2.0 + Fy**2.0))

    # Calculate sigma to eventuate given bandwidth (in octaves)
    sigma = (
        center_frequency
        / ((2.0**bandwidth + 1) * np.sqrt(2.0 * np.log(2.0)))
        * (2.0**bandwidth - 1)
    )

    # Create bandpass filter
    fil = 1.0 / (np.sqrt(2.0 * np.pi) * sigma) * np.exp(-(distance**2.0) / (2.0 * sigma**2.0))
    fil = fil / fil.max()

    stim = {
        "img": fil,
        "visual_size": visual_size,
        "ppd": ppd,
        "shape": shape,
        "center_frequency": center_frequency,
        "sigma": sigma,
        "frequency_extent": [fy[0], fy[-1], fx[0], fx[-1]],
    }

    return stim



import itertools

import numpy as np

from stimupy.components import *  # angulars, edges, frames, gaussians, lines, radials, shapes, waves
from stimupy.components import texts
from stimupy.utils import resolution

__all__ = [
    "overview",
    "plot_overview",
    "image_base",
    "draw_regions",
    "mask_regions",
    "combine_masks",
    "overview",
    "angulars",
    "radials",
    "edges",
    "frames",
    "gaussians",
    "lines",
    "shapes",
    "texts",
    "waves",
]


def image_base(visual_size=None, shape=None, ppd=None, rotation=0.0, origin="mean"):
    """Create coordinate-arrays to serve as image base for drawing

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    rotation : float, optional
        rotation (in degrees) from 3 o'clock, counterclockwise, by default 0.0
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with keys:
        "visual_size", "ppd" : resolved from input arguments,
        "x", "y" : single axes
        "horizontal", "vertical" : numpy.ndarray of shape, with distance from origin,
        in deg. visual angle, at each pixel
        "oblique", "oblique_y" : numpy.ndarray of shape, with oblique distances from origin,
        in deg. visual angle, at each pixel
        "radial" : numpyn.ndarray of shape, with radius from origin,
        in deg. visual angle, at each pixel
        "angular" : numpy.ndarray of shape, with angle relative to 3 o'clock,
        in rad, at each pixel
        "rectilinear" : numpy.ndarray of shape, with rectilinear/cityblock/Manhattan distance from origin,
        in deg. visual angle, at each pixel
    """

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    # Get single axes
    x, y = resolution.visual_size_to_axes(visual_size=visual_size, shape=shape, origin=origin)

    # Linear distance image bases
    xx, yy = np.meshgrid(x, y)

    # Rotate to get obliques
    alpha = [np.cos(np.deg2rad(-rotation)), np.sin(np.deg2rad(-rotation))]
    beta = [np.cos(np.deg2rad(rotation)), np.sin(np.deg2rad(rotation))]
    oblique_x = alpha[0] * xx + alpha[1] * yy
    oblique_y = beta[1] * xx + beta[0] * yy
    if origin == "corner":
        oblique_x = oblique_x - oblique_x.min()
        oblique_y = oblique_y - oblique_y.min()

    # Rectilinear distance (frames)
    rectilinear = np.maximum(np.abs(oblique_x), np.abs(oblique_y))

    # Radial distance
    radial = np.sqrt(xx**2 + yy**2)

    # Angular distance
    angular = np.arctan2(xx, yy)
    angular -= np.deg2rad(rotation + 90)
    angular %= 2 * np.pi

    return {
        "visual_size": visual_size,
        "ppd": ppd,
        "shape": shape,
        "rotation": rotation,
        "x": x,
        "y": y,
        "horizontal": xx,
        "vertical": yy,
        "oblique": oblique_x,
        "oblique_y": oblique_y,
        "rectilinear": rectilinear,
        "radial": radial,
        "angular": angular,
    }


def mask_regions(
    distance_metric,
    edges,
    shape=None,
    visual_size=None,
    ppd=None,
    rotation=0.0,
    origin=None,
):
    """Generate mask for regions in image

    Regions are defined by `edges` along a `distance_metric`.
    Regions will be masked consecutively, from `origin` outwards,
    such that each `edge` is the upper-limit of a region.

    Parameters
    ----------
    distance_metric : any of keys in stimupy.components.image_base()
        which distance metric to mask over
    edges : Sequence[Number]
        upper-limit of each consecutive region
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    rotation : float, optional
        rotation (in degrees) from 3 o'clock, counterclockwise, by default 0.0
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        mask with integer index for each angular segment (key: "mask"),
        and additional keys containing stimulus parameters
    """

    # Set up coordinates
    base = image_base(
        shape=shape, visual_size=visual_size, ppd=ppd, rotation=rotation, origin=origin
    )
    distances = base[distance_metric]
    distances = np.round(distances, 8)

    if isinstance(edges, (int, float)):
        edges = (edges,)

    # Mark elements with integer idx-value
    mask = np.zeros(base["shape"], dtype=int)
    for idx, edge in zip(reversed(range(len(edges))), reversed(edges)):
        mask[distances <= edge] = int(idx + 1)

    # Assemble output
    return {
        "mask": mask,
        "edges": edges,
        "distance_metric": distance_metric,
        "rotation": base["rotation"],
        "shape": base["shape"],
        "visual_size": base["visual_size"],
        "ppd": base["ppd"],
        "distances": distances,
        "origin": origin,
    }


def combine_masks(*masks):
    """Combines several masks into a singular mask

    Increments mask-indices, such that the resulting mask contains consecutive integer
    indices.
    Masks are combined in order.

    Parameters
    ----------
    mask_1, mask_2, ... : numpy.ndarray
        Masks to be combined

    Returns
    -------
    numpy.ndarray
        Combined mask, where integer indices are in order of the input masks.

    Raises
    ------
    ValueError
        if masks do not all have the same shape (in pixels)
    ValueError
        if multiple masks index the same pixel
    """
    # Initialize
    combined_mask = np.zeros_like(masks[0])
    for mask in masks:
        # Check that masks have the same shape
        if not mask.shape == combined_mask.shape:
            raise ValueError("Not all masks have the same shape")

        # Check that masks don't overlap
        if (combined_mask & mask).any():
            raise ValueError("Masks overlap")

        # Combine: increase `mask`-idc by adding the current highest idx in combined_mask
        combined_mask = np.where(mask, mask + combined_mask.max(), combined_mask)

    return combined_mask


def draw_regions(mask, intensities, intensity_background=0.5):
    """Draw regions defined by mask, with given intensities

    Parameters
    ----------
    mask : numpy.ndarray
        image-array with integer-indices for each region to draw
    intensities : Sequence[float, ...]
        intensity value for each masked region.
        Can specify as many intensities as number of masked regions;
        If fewer intensities are passed than masked regions, cycles through intensities
    intensity_background : float, optional
        intensity value of background, by default 0.5

    Returns
    -------
    numpy.ndarray
        image-array, same shape as mask, with intensity assigned to each masked region
    """

    # Create background
    img = np.ones(mask.shape) * intensity_background

    # Get mask indices
    mask_idcs = np.unique(mask[mask > 0])

    if isinstance(intensities, (float, int)):
        intensities = (intensities,)

    # Assign intensities to masked regions
    ints = [*itertools.islice(itertools.cycle(intensities), len(mask_idcs))]
    for frame_idx, intensity in zip(mask_idcs, ints):
        img = np.where(mask == frame_idx, intensity, img)

    return img


def overview(skip=False):
    """Generate example stimuli from this module

    Returns
    -------
    dict[str, dict]
        Dict mapping names to individual stimulus dicts
    """
    stimuli = {}
    for stimmodule_name in __all__:
        if stimmodule_name in [
            "overview",
            "plot_overview",
            "draw_regions",
            "image_base",
            "mask_regions",
            "combine_masks",
        ]:
            continue

        print(f"Generating stimuli from {stimmodule_name}")
        # Get a reference to the actual module
        stimmodule = globals()[stimmodule_name]
        try:
            stims = stimmodule.overview()

            # Accumulate
            stimuli.update(stims)
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stimuli


def plot_overview(mask=False, save=None, units="deg"):
    """Plot overview of examples in this module (and submodules)

    Parameters
    ----------
    mask : bool or str, optional
        If True, plot mask on top of stimulus image (default: False).
        If string is provided, plot this key from stimulus dictionary as mask
    save : None or str, optional
        If None (default), do not save the plot.
        If string is provided, save plot under this name.
    units : "px", "deg" (default), or str
        what units to put on the axes, by default degrees visual angle ("deg").
        If a str other than "deg"(/"degrees") or "px"(/"pix"/"pixels") is passed,
        it must be the key to a tuple in stim

    """
    from stimupy.utils import plot_stimuli

    stims = overview(skip=True)
    plot_stimuli(stims, mask=mask, units=units, save=save)


if __name__ == "__main__":
    plot_overview()



import warnings

import numpy as np

from stimupy.components import draw_regions, waves
from stimupy.utils import resolution
from stimupy.utils.contrast_conversions import transparency

__all__ = [
    "checkerboard",
    "contrast_contrast",
]


def mask_from_idx(checkerboard_stim, check_idc):
    """Create binary mask for specified check indidces

    Parameters
    ----------
    checkerboard_stim : dict
        stimulus dictionary of checkerboard,
    check_idc : Sequence[(Number, Number),...]
        target indices (row, column of checkerboard) of checks to create mask for

    Returns
    -------
    numpy.ndarray
        mask, as binary 2D numpy.ndarray with 1 for all pixels belonging to
        specified check(s), and 0 everywhere else

    Raises
    ------
    ValueError
        Check index is invalid given the board shape
    """
    board_shape = checkerboard_stim["board_shape"]
    mask = np.zeros(checkerboard_stim["shape"])
    for i, coords in enumerate(check_idc):
        if coords[0] < 0 or coords[0] > coords[0] or coords[1] < 0 or coords[1] > board_shape[1]:
            raise ValueError(f"Cannot provide mask for check {coords} outside board {board_shape}")
        m1 = np.where(checkerboard_stim["col_mask"] == coords[1] + 1, 1, 0)
        m2 = np.where(checkerboard_stim["row_mask"] == coords[0] + 1, 1, 0)
        mask = np.where(m1 + m2 == 2, i + 1, mask)

        if len(np.unique(mask)) == 1:
            raise ValueError(
                f"Cannot provide mask for check {coords} outside board because of rotation"
            )
    return mask


def extend_target_idx(target_index, offsets=[(-1, 0), (0, -1), (0, 0), (0, 1), (1, 0)]):
    """Extend target indices, to not just the specified check(s) but also surrounding

    Parameters
    ----------
    target_index : (Number, Number)
        target indices (row, column of checkerboard)
    offsets : list, optional
        relative indices of neighboring checks to include in target,
        by default [(-1, 0), (0, -1), (0, 0), (0, 1), (1, 0)]

    Returns
    -------
    List[Tuple(Number, Number),...]
        List of all target indices
    """
    extended_idc = []
    for offset in offsets:
        new_idx = (target_index[0] + offset[0], target_index[1] + offset[1])
        extended_idc.append(new_idx)
    return extended_idc


def add_targets(checkerboard_stim, target_indices=(), extend_targets=False, intensity_target=0.5):
    """Add targets to a checkerboard stimulus

    Parameters
    ----------
    checkerboard_stim : dict
        stimulus dictionary of checkerboard,
        needs to contain at least "img" and "board_shape"
    target_indices : Sequence[(Number, Number),...]
        target indices (row, column of checkerboard)
    extend_targets : bool, optional
        if true, extends the targets by 1 check in all 4 directions, by default False
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities

    Returns
    -------
    dict[str, Any]
        dict with the updated stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    if target_indices is None:
        target_indices = ()
    target_mask = np.zeros(checkerboard_stim["shape"])

    # Define target mask
    for i, target in enumerate(target_indices):
        if extend_targets:
            target_idc = extend_target_idx(target)
        else:
            target_idc = [target]
        for target_idx in target_idc:
            target_mask += mask_from_idx(checkerboard_stim, (target_idx,)) * (i + 1)
    checkerboard_stim["target_mask"] = target_mask.astype(int)

    # Draw targets
    checkerboard_stim["img"] = np.where(
        target_mask,
        draw_regions(mask=target_mask, intensities=intensity_target, intensity_background=0.0),
        checkerboard_stim["img"],
    )

    return checkerboard_stim


def checkerboard(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    board_shape=None,
    check_visual_size=None,
    target_indices=(),
    extend_targets=False,
    period="ignore",
    rotation=0.0,
    intensity_checks=(0.0, 1.0),
    intensity_target=0.5,
    round_phase_width=True,
):
    """Checkerboard assimilation effect

    High-contrast checkerboard, with intermediate targets embedded in it.
    Target brightness assimilates to direct surround, rather than contrast with it.

    These kinds of checkerboard displays are described by De Valois & De Valois (1988),
    and the brightness effect of it by Blakeslee & McCourt (2004).

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size of the total board [height, width] in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] in pixels
    frequency : Sequence[Number, Number], Number, or None (default)
        frequency of checkerboard in [y, x] in cpd
    board_shape : Sequence[Number, Number], Number, or None (default)
        number of checks in [height, width] of checkerboard
    check_visual_size : Sequence[Number, Number], Number, or None (default)
        visual size of a single check [height, width] in degrees
    target_indices : Sequence[(Number, Number),...], optional
        target indices (row, column of checkerboard), by default None
    extend_targets : bool, optional
        if true, extends the targets by 1 check in all 4 directions, by default False
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_checks : Sequence[float, float]
        intensity values of checks, by default (0.0, 1.0)
    round_phase_width : Bool
        if True, round width of bars given resolution (default: True)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2004).
        A unified theory of brightness contrast and assimilation
        incorporating oriented multiscale spatial filtering and contrast normalization.
        Vision Research, 44(21), 2483-2503.
        https://doi.org/10/fmcx5p
    De Valois, R. L., & De Valois, K. K. (1988).
        Spatial Vision. Oxford University Press.
    """

    lst = [visual_size, ppd, shape, frequency, board_shape, check_visual_size]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'checkerboard()' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'board_shape', 'check_visual_size'"
        )

    if isinstance(frequency, (float, int)) or frequency is None:
        frequency = (frequency, frequency)
    if isinstance(board_shape, (float, int)) or board_shape is None:
        board_shape = (board_shape, board_shape)
    if isinstance(check_visual_size, (float, int)) or check_visual_size is None:
        check_visual_size = (check_visual_size, check_visual_size)

    create_twice = visual_size is None and shape is None

    # Create checkerboard by treating it as a plaid
    sw1 = waves.square(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency[0],
        n_phases=board_shape[0],
        phase_width=check_visual_size[0],
        period=period,
        rotation=rotation,
        phase_shift=0,
        intensities=intensity_checks,
        origin="corner",
        round_phase_width=round_phase_width,
        distance_metric="oblique",
    )

    sw2 = waves.square(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency[1],
        n_phases=board_shape[1],
        phase_width=check_visual_size[1],
        period=period,
        rotation=rotation - 90,
        phase_shift=0,
        intensities=intensity_checks,
        origin="corner",
        round_phase_width=round_phase_width,
        distance_metric="oblique",
    )

    # If neither a visual_size nor a shape was given, each square wave
    # grating is always a square. An easy solution is to just recreate
    # both gratings with the resolved parameters
    if create_twice:
        warnings.filterwarnings("ignore")
        sw1 = waves.square(
            visual_size=(sw1["visual_size"][0], sw2["visual_size"][1]),
            ppd=sw1["ppd"],
            shape=None,
            frequency=frequency[0],
            n_phases=board_shape[0],
            phase_width=check_visual_size[0],
            period=period,
            rotation=rotation,
            phase_shift=0,
            intensities=intensity_checks,
            origin="corner",
            round_phase_width=round_phase_width,
            distance_metric="oblique",
        )

        sw2 = waves.square(
            visual_size=(sw1["visual_size"][0], sw2["visual_size"][1]),
            ppd=sw1["ppd"],
            shape=None,
            frequency=frequency[1],
            n_phases=board_shape[1],
            phase_width=check_visual_size[1],
            period=period,
            rotation=rotation - 90,
            phase_shift=0,
            intensities=intensity_checks,
            origin="corner",
            round_phase_width=round_phase_width,
            distance_metric="oblique",
        )
        warnings.filterwarnings("default")

    # Add the two square-wave gratings into a checkerboard
    img = sw1["img"] + sw2["img"]
    img = np.where(
        img == intensity_checks[0] + intensity_checks[1], intensity_checks[1], intensity_checks[0]
    )

    # Create a mask with target indices for each check
    mask = sw1["grating_mask"] + sw2["grating_mask"] * sw1["grating_mask"].max() * 10
    unique_vals = np.unique(mask)
    for v in range(len(unique_vals)):
        mask[mask == unique_vals[v]] = v + 1

    stim = {
        "img": img,
        "checker_mask": mask.astype(int),
        "col_mask": sw1["grating_mask"],
        "row_mask": sw2["grating_mask"],
        "visual_size": sw1["visual_size"],
        "ppd": sw1["ppd"],
        "shape": sw1["shape"],
        "frequency": (sw2["frequency"], sw1["frequency"]),
        "board_shape": (sw2["n_phases"], sw1["n_phases"]),
        "check_visual_size": (sw2["phase_width"], sw1["phase_width"]),
        "period": period,
        "rotation": rotation,
        "intensity_checks": intensity_checks,
        "target_indices": target_indices,
        "extend_targets": extend_targets,
        "round_phase_width": round_phase_width,
    }

    # Add targets
    stim = add_targets(
        stim,
        target_indices=target_indices,
        extend_targets=extend_targets,
        intensity_target=intensity_target,
    )

    return stim


def contrast_contrast(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    board_shape=None,
    check_visual_size=None,
    target_shape=None,
    period="ignore",
    rotation=0.0,
    intensity_checks=(0.0, 1.0),
    tau=0.5,
    alpha=None,
    round_phase_width=True,
):
    """Contrast-contrast effect on checkerboard with square transparency layer

    Checkerboard version of the contrast-contrast illusion (Chubb, Sperling, Solomon,
    1989), as used by Domijan (2015).

    Parameters
    ----------
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] in pixels
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size of the total board [height, width] in degrees
    board_shape : Sequence[Number, Number], Number, or None (default)
        number of checks in [height, width] of checkerboard
    check_visual_size : Sequence[Number, Number], Number, or None (default)
        visual size of a single check [height, width] in degrees
    targets_shape : Sequence[Number, Number], Number, or None (default)
        number of checks with transparency in y, x direction
    intensity_low : float, optional
        intensity value of the dark checks, by default 0.0
    intensity_high : float, optional
        intensity value of the light checks, by default 1.0
    tau : Number
        tau of transparency (i.e. value of transparent medium), default 0.5
    alpha : Number or None (default)
        alpha of transparency (i.e. how transparent the medium is)
    round_phase_width : Bool
        if True, round width of bars given resolution (default: True)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Chubb, C., Sperling, G., & Solomon, J. A. (1989).
        Texture interactions determine perceived contrast.
        Proc. Natl. Acad. Sci. USA, 5.
        https://doi.org/10.1073/pnas.86.23.9631
    Domijan, D. (2015).
        A Neurocomputational account of the role
        of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9(February), 1-16.
        https://doi.org/10/gh62x2
    """
    if target_shape is None:
        raise ValueError("contrast_contrast() missing argument 'target_shape' which is not 'None'")
    if alpha is None:
        raise ValueError("contrast_contrast() missing argument 'alpha' which is not 'None'")
    if isinstance(target_shape, (int, float)):
        target_shape = (target_shape, target_shape)

    # Set up basic checkerboard
    stim = checkerboard(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        board_shape=board_shape,
        check_visual_size=check_visual_size,
        period=period,
        rotation=rotation,
        intensity_checks=intensity_checks,
        round_phase_width=round_phase_width,
    )
    img = stim["img"]

    # Determine target locations
    check_size_px = resolution.shape_from_visual_size_ppd(
        visual_size=stim["check_visual_size"], ppd=stim["ppd"]
    )
    target_idx = np.zeros(img.shape, dtype=bool)
    tposy = (img.shape[0] - target_shape[0] * check_size_px.height) // 2
    tposx = (img.shape[1] - target_shape[1] * check_size_px.width) // 2
    target_idx[
        tposy : tposy + target_shape[0] * check_size_px[0],
        tposx : tposx + target_shape[1] * check_size_px[1],
    ] = True

    # Construct mask for target region
    mask = np.zeros(img.shape)
    mask[target_idx] = 1

    # Apply transparency to target locations
    img = transparency(img=img, mask=mask, alpha=alpha, tau=tau)

    stim["img"] = img
    stim["target_mask"] = mask.astype(int)
    stim["target_shape"] = target_shape
    stim["alpha"] = alpha
    stim["tau"] = tau
    stim["round_phase_width"] = round_phase_width
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {"visual_size": (10, 10), "ppd": 30}
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "checkerboard": checkerboard(**default_params, check_visual_size=(1, 1)),
        "checkerboard_from_frequency": checkerboard(**default_params, frequency=1, rotation=45),
        "checkerboard_with_targets": checkerboard(**default_params, check_visual_size=(1, 1), target_indices=[(3, 2), (5, 5)]),
        "checkerboard_contrast_contrast": contrast_contrast(**default_params, check_visual_size=(1, 1), target_shape=4, alpha=0.2),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import copy
import warnings

import numpy as np
from PIL import Image, ImageDraw

from stimupy.components.shapes import ellipse as ellipse_shape
from stimupy.utils import resolution

__all__ = [
    "line",
    "dipole",
    "ellipse",
    "circle",
]


def line(
    visual_size=None,
    ppd=None,
    shape=None,
    line_position=None,
    line_length=None,
    line_width=0,
    rotation=0.0,
    intensity_line=1.0,
    intensity_background=0.0,
    origin="corner",
):
    """Draw a line

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    line_position : Sequence[Number, Number], Number, or None (default)
        line position (y, x) given the chosen origin;
        if None (default), the lines will go through the image center
    line_length : Number
        length of the line, in degrees visual angle
    line_width : Number
        width of the line, in degrees visual angle;
        if line_width=0 (default), line will be one pixel wide
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_line : Number
        intensity value of the line (default: 1)
    intensity_background : Number
        intensity value of the background (default: 0)
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean" or "center": set origin to center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each line (key: "line_mask"),
        and additional keys containing stimulus parameters
    """
    if line_length is None:
        raise ValueError("line() missing argument 'line_length' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    alpha = [np.cos(np.deg2rad(rotation)), np.sin(np.deg2rad(rotation))]

    if isinstance(line_position, (float, int)) and line_position is not None:
        line_position = (line_position, line_position)

    if line_position is None:
        origin = "center"
        line_position = (-line_length * alpha[0] / 2, -line_length * alpha[1] / 2)

    if origin == "corner":
        position = (line_position[0] * ppd[0], line_position[1] * ppd[1])
    elif origin == "center" or "mean":
        position = (
            int(np.round(line_position[0] * ppd[0] + shape[0] / 2)),
            int(np.round(line_position[1] * ppd[1] + shape[1] / 2)),
        )
    else:
        raise ValueError("origin must be corner, center or mean")

    line_width_old = copy.deepcopy(line_width)
    line_width = np.round(line_width * ppd[0]) / ppd[0]
    if line_width != line_width:
        warnings.warn(f"Rounding line_width; {line_width_old} -> {line_width}")
    if line_width == 0:
        warnings.warn("line_width == 0 -> using line_width of 1px")

    # Create Pillow Image object
    img = Image.new("RGB", (shape.width, shape.height))

    # Calculate line coordinates
    coords = (
        position[::-1],
        (
            int(np.round(position[1] + line_length * alpha[1] * ppd[1])),
            int(np.round(position[0] + line_length * alpha[0] * ppd[0])),
        ),
    )

    # Create line image
    ImageDraw.Draw(img).line(coords, width=int(line_width * ppd[0]))

    # Convert to numpy array, create mask and adapt intensities
    img = np.array(img)[:, :, 0] / 255
    mask = copy.deepcopy(img)
    img = img * (intensity_line - intensity_background) + intensity_background

    stim = {
        "img": img,
        "line_mask": mask.astype(int),
        "visual_size": visual_size,
        "ppd": ppd,
        "shape": shape,
        "line_position": line_position,
        "line_length": line_length,
        "line_width": line_width,
        "rotation": rotation,
        "intensity_line": intensity_line,
        "intensity_background": intensity_background,
        "origin": origin,
    }
    return stim


def dipole(
    visual_size=None,
    ppd=None,
    shape=None,
    line_length=None,
    line_width=0,
    line_gap=None,
    rotation=0.0,
    intensity_lines=(0.0, 1.0),
):
    """Draw a two centered parallel lines

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    line_length : Number
        length of the line, in degrees visual angle
    line_width : Number
        width of the line, in degrees visual angle;
        if line_width=0 (default), line will be one pixel wide
    line_gap : Number
        distance between line centers, in degrees visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_lines : (Number, Number)
        intensity value of the line (default: (0, 1));
        background intensity is the mean of these two values

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each line (key: "line_mask"),
        and additional keys containing stimulus parameters
    """
    if line_length is None:
        raise ValueError("dipole() missing argument 'line_length' which is not 'None'")
    if line_gap is None:
        raise ValueError("dipole() missing argument 'line_gap' which is not 'None'")
    if line_gap == 0:
        raise ValueError("line_gap should be larger than 0")

    intensity_background = (intensity_lines[0] + intensity_lines[1]) / 2
    alpha1 = [np.cos(np.deg2rad(rotation)), np.sin(np.deg2rad(rotation))]
    alpha2 = [np.cos(np.deg2rad(rotation + 90)), np.sin(np.deg2rad(rotation + 90))]

    line_position1 = (
        -line_length * alpha1[0] / 2 + line_gap / 2 * alpha2[0],
        -line_length * alpha1[1] / 2 + line_gap / 2 * alpha2[1],
    )

    line_position2 = (
        -line_length * alpha1[0] / 2 - line_gap / 2 * alpha2[0],
        -line_length * alpha1[1] / 2 - line_gap / 2 * alpha2[1],
    )

    stim1 = line(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        line_position=line_position1,
        line_length=line_length,
        line_width=line_width,
        rotation=rotation,
        intensity_line=intensity_lines[0],
        intensity_background=intensity_background,
        origin="center",
    )

    stim2 = line(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        line_position=line_position2,
        line_length=line_length,
        line_width=line_width,
        rotation=rotation,
        intensity_line=intensity_lines[1] - intensity_background,
        intensity_background=0,
        origin="center",
    )

    stim1["img"] = stim1["img"] + stim2["img"]
    stim1["line_mask"] = (stim1["line_mask"] + stim2["line_mask"] * 2).astype(int)
    stim1["line_gap"] = line_gap
    stim1["intensity_lines"] = intensity_lines
    del stim1["intensity_line"]

    if line_width == 0:
        line_width = 1 / np.unique(stim1["ppd"])
    if line_width >= line_gap:
        raise ValueError("line_width should not be larger than line_gap")

    return stim1


def ellipse(
    visual_size=None,
    ppd=None,
    shape=None,
    radius=None,
    line_width=0,
    intensity_line=1.0,
    intensity_background=0.0,
):
    """Draw an ellipse

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radius : Sequence[Number, Number], Number or None (default)
        ellipse radius [ry, rx] in degrees visual angle
    line_width : Number
        width of the line, in degrees visual angle;
        if line_width=0 (default), line will be one pixel wide
    intensity_line : Number
        intensity value of the line (default: 1)
    intensity_background : Number
        intensity value of the background (default: 0)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each line (key: "line_mask"),
        and additional keys containing stimulus parameters
    """
    if radius is None:
        raise ValueError("ellipse() missing argument 'radius' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if line_width * ppd[0] == 0:
        line_width_ = 1 / ppd[0]
    else:
        line_width_ = line_width

    stim = ellipse_shape(
        radius=np.array(radius),
        intensity_ellipse=intensity_line,
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        intensity_background=intensity_background,
        origin="mean",
    )

    stim2 = ellipse_shape(
        radius=np.array(radius) - line_width_,
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        origin="mean",
    )

    stim["img"] = np.where(stim2["ellipse_mask"] == 1, intensity_background, stim["img"])
    stim["line_mask"] = np.where(stim2["ellipse_mask"] == 1, 0, stim["ellipse_mask"])
    stim["intensity_line"] = intensity_line
    stim["line_width"] = line_width
    del stim["ellipse_mask"], stim["intensity_ellipse"]
    return stim


def circle(
    visual_size=None,
    ppd=None,
    shape=None,
    radius=None,
    line_width=0,
    intensity_line=1.0,
    intensity_background=0.0,
):
    """Draw a circle given the input parameters

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radius : Number
        radius of circle in degrees visual angle
    line_width : Number
        width of the line, in degrees visual angle;
        if line_width=0 (default), line will be one pixel wide
    intensity_line : Number
        intensity value of the line (default: 1)
    intensity_background : Number
        intensity value of the background (default: 0)

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each line (key: "line_mask"),
        and additional keys containing stimulus parameters
    """
    if radius is None:
        raise ValueError("circle() missing argument 'radius' which is not 'None'")
    if not isinstance(radius, (int, float)):
        raise ValueError("radius should be a single number")

    stim = ellipse(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        radius=radius,
        line_width=line_width,
        intensity_line=intensity_line,
        intensity_background=intensity_background,
    )
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 10,
    }
    default_params.update(kwargs)

    p = {
        "line_length": 2,
        "line_width": 0.01,
        "rotation": 30,
    }

    # fmt: off
    stimuli = {
        "lines_line": line(**default_params, **p, origin="center"),
        "lines_dipole": dipole(**default_params, **p, line_gap=1),
        "lines_circle": circle(**default_params, radius=3),
        "lines_ellipse": ellipse(**default_params, radius=(3, 4)),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)
